#ifndef _STDRAND
#define _STDRAND

#include "xoroshiro128.hpp"

#include <random>

namespace stdrand
{
    ///////////////////////////////////////////////////////////////////////////
    /// Base class for the random number generators.
    ///////////////////////////////////////////////////////////////////////////
    class std_randbase
    {
    protected:
        rng::rng128 generator;
        // std::mt19937_64 generator;
    public:
    	////////////////////////////////////////////////////////////////////////
    	/// Default constructor.
    	////////////////////////////////////////////////////////////////////////
    	std_randbase(){}
    	////////////////////////////////////////////////////////////////////////
    	/// Default destructor.
    	////////////////////////////////////////////////////////////////////////
    	~std_randbase(){}
    	////////////////////////////////////////////////////////////////////////
    	/// \brief Change the current random seed.
    	///
    	/// \param seed The new seed.
    	////////////////////////////////////////////////////////////////////////
    	void change_seed(int seed)
        {
            generator = rng::rng128(seed);
            // generator.seed(seed);
        }
        ////////////////////////////////////////////////////////////////////////
    	/// Return a single random number.
    	////////////////////////////////////////////////////////////////////////
    	virtual double gen(){return 0;}
        ////////////////////////////////////////////////////////////////////////
    	/// \brief Jump forward 2^64 places
    	////////////////////////////////////////////////////////////////////////
    	void jump()
        {
            generator.jump();
            // for(int i=0; i < 1e6; i++)
            // {this->gen();}
        }
    };

    ///////////////////////////////////////////////////////////////////////////
    /// Generator for uniform numbers between 0 and 1
    ///////////////////////////////////////////////////////////////////////////
    class std_d_unirand : public std_randbase
    {
    private:
        std::uniform_real_distribution<double> distribution;
    public:
        ////////////////////////////////////////////////////////////////////////
    	/// \brief Constructor
    	///
    	/// \param seed The inital seed of the random number generator.
    	////////////////////////////////////////////////////////////////////////
        std_d_unirand(int seed);
        ////////////////////////////////////////////////////////////////////////
    	/// \brief Constructor with dummy size input
    	///
    	/// \param seed The inital seed of the random number generator.
        /// \param size A dummy size input for compatibility with Intel RNG.
    	////////////////////////////////////////////////////////////////////////
        std_d_unirand(int size, int seed) : std_d_unirand(seed) {}
        ////////////////////////////////////////////////////////////////////////
    	/// Default destructor.
    	////////////////////////////////////////////////////////////////////////
    	~std_d_unirand(){}
        ////////////////////////////////////////////////////////////////////////
    	/// Return a single random number.
    	////////////////////////////////////////////////////////////////////////
    	double gen(){return distribution(generator);}
    };

    ///////////////////////////////////////////////////////////////////////////
    /// Generator for uniform integers between 0 and 1
    ///////////////////////////////////////////////////////////////////////////
    class std_i_unirand : public std_randbase
    {
    private:
        std::uniform_int_distribution<int> distribution;
    public:
        ////////////////////////////////////////////////////////////////////////
    	/// \brief Constructor
    	///
    	/// \param seed The inital seed of the random number generator.
    	////////////////////////////////////////////////////////////////////////
        std_i_unirand(int seed);
        ////////////////////////////////////////////////////////////////////////
    	/// \brief Constructor with dummy size input
    	///
    	/// \param seed The inital seed of the random number generator.
        /// \param size A dummy size input for compatibility with Intel RNG.
    	////////////////////////////////////////////////////////////////////////
        std_i_unirand(int size, int seed) : std_i_unirand(seed=seed) {}
        ////////////////////////////////////////////////////////////////////////
    	/// Default destructor.
    	////////////////////////////////////////////////////////////////////////
    	~std_i_unirand(){}
        ////////////////////////////////////////////////////////////////////////
    	/// Return a single random number.
    	////////////////////////////////////////////////////////////////////////
    	double gen(){return distribution(generator);}
    };

    ///////////////////////////////////////////////////////////////////////////
    /// Generator for random numbers on a normal distribution
    ///////////////////////////////////////////////////////////////////////////
    class std_normrand : public std_randbase
    {
    private:
        std::normal_distribution<double> distribution;
    public:
        ////////////////////////////////////////////////////////////////////////
    	/// \brief Constructor
    	///
    	/// \param m The mean of the normal distribution.
    	/// \param sdin The standard deviation of the normal distribution.
    	/// \param seed The inital seed of the random number generator. Defaults
    	///             to 1.
    	////////////////////////////////////////////////////////////////////////
        std_normrand(double m, double sdin, int seed=1);
        ////////////////////////////////////////////////////////////////////////
    	/// Default destructor.
    	////////////////////////////////////////////////////////////////////////
    	~std_normrand(){}
        ////////////////////////////////////////////////////////////////////////
    	/// Return a single random number.
    	////////////////////////////////////////////////////////////////////////
    	double gen(){return distribution(generator);}
    };

    ///////////////////////////////////////////////////////////////////////////
    /// Generator for random numbers on a lognormal distribution
    ///////////////////////////////////////////////////////////////////////////
    class std_lognormrand : public std_randbase
    {
    private:
        std::lognormal_distribution<double> distribution;
    public:
        ////////////////////////////////////////////////////////////////////////
    	/// \brief Constructor
    	///
    	/// \param m The logarithmic mean of the lognormal distribution.
    	/// \param sdin The logarithmic standard deviation of the lognormal
        ///             distribution.
    	/// \param seed The inital seed of the random number generator.
    	////////////////////////////////////////////////////////////////////////
        std_lognormrand(double m, double sdin, int seed);
        ////////////////////////////////////////////////////////////////////////
    	/// \brief Constructor with dummy size
    	///
    	/// \param m The logarithmic mean of the lognormal distribution.
    	/// \param sdin The logarithmic standard deviation of the lognormal
        ///             distribution.
    	/// \param seed The inital seed of the random number generator.
        /// \param size A dummy size input for compatibility with Intel RNG.
    	////////////////////////////////////////////////////////////////////////
        std_lognormrand(double m, double sdin, int size, int seed)
            : std_lognormrand(m, sdin, seed) {}
        ////////////////////////////////////////////////////////////////////////
    	/// Default destructor.
    	////////////////////////////////////////////////////////////////////////
    	~std_lognormrand(){}
        ////////////////////////////////////////////////////////////////////////
    	/// Return a single random number.
    	////////////////////////////////////////////////////////////////////////
    	double gen(){return distribution(generator);}
    };
}

#endif
